﻿using System.Collections.Generic;
using System.Collections;
using System;

namespace Persistence {

  public class Serializer : IEnumerable<ISerializationToken> {

    private object obj;
    private string rootName;
    private Dictionary<object, int> obj2id;
    private Dictionary<Type, IPrimitiveSerializer> primitiveSerializers;
    private List<ICompoundSerializer> compoundSerializers;

    public Serializer(object obj,
        IEnumerable<IPrimitiveSerializer> primitiveSerializers,
        IEnumerable<ICompoundSerializer> compoundSerializers) :
      this(obj, primitiveSerializers, compoundSerializers, "ROOT") { }

    public Serializer(object obj,
        IEnumerable<IPrimitiveSerializer> primitiveSerializers,
        IEnumerable<ICompoundSerializer> compoundSerializers, string rootName) {
      this.obj = obj;
      this.rootName = rootName;
      this.primitiveSerializers = new Dictionary<Type, IPrimitiveSerializer>();
      foreach (IPrimitiveSerializer serializer in primitiveSerializers) {
        this.primitiveSerializers.Add(serializer.Type, serializer);
      }
      this.compoundSerializers = new List<ICompoundSerializer>(compoundSerializers);
      this.obj2id = new Dictionary<object, int>();
      obj2id.Add(new object(), 0);
    }

    IEnumerator IEnumerable.GetEnumerator() {
      return this.GetEnumerator();
    }

    public IEnumerator<ISerializationToken> GetEnumerator() {
      DataMemberAccessor rootAccessor = new DataMemberAccessor(
        this.rootName, obj.GetType(), null, () => this.obj, null);
      IEnumerator<ISerializationToken> iterator = Serialize(rootAccessor);
      while (iterator.MoveNext())
        yield return iterator.Current;
    }

    private IEnumerator<ISerializationToken> Serialize(DataMemberAccessor accessor) {
      object value = accessor.Get();
      if (value == null) {
        yield return new NullReferenceToken(accessor.Name);
      } else if (this.primitiveSerializers.ContainsKey(value.GetType())) {
        yield return new PrimitiveToken(accessor, this.primitiveSerializers[value.GetType()].Serialize(value));
      } else if (this.obj2id.ContainsKey(value)) {
        yield return new ReferenceToken(accessor.Name, this.obj2id[value]);
      } else {
        int id = obj2id.Count;
        this.obj2id.Add(value, id);
        yield return new BeginToken(accessor, id);
        ICompoundSerializer customSerializer = this.FindCompoundSerializer(value.GetType());
        if (customSerializer != null) {
          foreach (object obj in customSerializer.Serialize(value)) {
            IEnumerator<ISerializationToken> iterator = this.Serialize(new DataMemberAccessor(obj));
            while (iterator.MoveNext())
              yield return iterator.Current;
          }
        } else { // composite serialization
          int nSubComponents = 0;
          foreach (KeyValuePair<string, DataMemberAccessor> mapping in
            StorableAttribute.GetAutostorableAccessors(value)) {
            nSubComponents += 1;
            IEnumerator<ISerializationToken> iterator = this.Serialize(mapping.Value);                        
            while (iterator.MoveNext()) {                            
              yield return iterator.Current;
            }                        
          }
          if (nSubComponents == 0) {
            throw new ApplicationException(String.Format(
              "Composite value of type \"{0}\" contains no subcomponents",
              value.GetType().FullName));
          }
        }
        yield return new EndToken(accessor, id);
      }
    }

    private ICompoundSerializer FindCompoundSerializer(Type type) {
      foreach (ICompoundSerializer s in compoundSerializers) {
        if (s.CanSerialize(type))
          return s;
      }
      return null;
    }
  }
}